import uuid
import datetime
import inspect
from typing import Type, TypeVar, Tuple, List, Optional, Dict

from sqlalchemy.sql import sqltypes

from fastapi_toolkit.generate.mock import MockMetadata, MockType
from fastapi_toolkit.define.link import Link


class BaseModel:
    def __init_subclass__(cls, **kwargs):
        register_model(cls)


def get_type_str(type) -> str:
    mod = inspect.getmodule(type)
    if mod is None:
        if type.__name__ == 'now':
            return 'datetime.datetime.now'
        raise TypeError("unknown type")
    mod = mod.__name__ + '.'
    if mod == 'builtins.':
        mod = ''
    if isinstance(type, sqltypes.String):
        return mod + f'String({type.length})'
    return mod + type.__name__


class Field:
    def __init__(self,
                 python_type,
                 sql_type,
                 primary_key: bool = False,
                 optional: bool = False,
                 default=None,
                 default_factory=None,
                 mock: Optional[MockMetadata] = None):
        self.python_type = python_type
        self.python_type_str = get_type_str(python_type)
        self.sql_type = sql_type
        self.sql_type_str = get_type_str(sql_type)
        self.primary_key = primary_key
        self.optional = optional
        if mock:
            self.mock = mock
        else:
            if python_type is int:
                self.mock = (MockType.int, (0, 100))
            elif python_type is float:
                self.mock = (MockType.float, (0, 100, 2))
            elif python_type is bool:
                self.mock = (MockType.bool, None)
            elif python_type is str:
                self.mock = (MockType.str, 10)
            elif python_type is uuid.UUID:
                self.mock = (MockType.uuid, None)
            elif python_type is datetime.datetime:
                self.mock = (MockType.datetime, None)
            else:
                self.mock = (MockType.str, 10)
        mock_type, mock_args = self.mock
        mock_func = f'mock_{mock_type.name}'
        if mock_args is not None:
            if isinstance(mock_args, tuple):
                mock_args = ', '.join([str(arg) for arg in mock_args])
            else:
                mock_args = str(mock_args)
            mock_args = f'{mock_args}'
        else:
            mock_args = ''
        self.mock = (mock_func, mock_args)
        assert not (default and default_factory), 'default and default_factory can not both exist'
        if default is not None:
            self.default = default
            self.default_str = str(default)
        if default_factory is not None:
            self.default_factory = default_factory
            self.default_factory_str = get_type_str(default_factory)


T = TypeVar('T', bound=BaseModel)


def register_model(model: Type[T]):
    info = ModelInfo()
    info.model = model
    for name, field in model.__dict__.items():
        if isinstance(field, Field):
            info.fields.append({'name': name, 'field': field})
    if hasattr(model, 'TKConfig'):
        config = model.TKConfig
        if hasattr(config, 'title'):
            info.title = config.title
        if hasattr(config, 'links'):
            info.links = config.links
            ModelManager.links.extend(config.links)
    ModelManager.models[model.__name__] = info


class ModelInfo:
    model: Type[T]
    fields: List[Dict[str, Field]]
    title: str = ''
    links: List[Link]

    def __init__(self):
        self.fields = []
        self.links = []


class ModelManager:
    models: Dict[str, ModelInfo] = {}
    links = []

    @staticmethod
    def _dont_appear_in_child(target: str, link: Link, chain: Optional[List[str]] = None) -> Tuple[bool, List]:
        if chain is None:
            chain = []
        _, nex = link.two_sides()
        chain.append(nex)
        if nex == target:
            return False, chain
        else:
            links = [l for l in ModelManager.links if nex == l.two_sides()[0]]
            for l in links:
                val, chain_ = ModelManager._dont_appear_in_child(target, l, chain.copy())
                if not val:
                    return False, chain_
            return True, []

    @staticmethod
    def check_links():
        # check link target is exist
        for link in ModelManager.links:
            a, b = link.two_sides()
            if a not in ModelManager.models:
                raise Exception(f"link's {a} not exist")
            if b not in ModelManager.models:
                raise Exception(f"link's {b} not exist")

        # check for circular links
        for link in ModelManager.links:
            a, b = link.two_sides()
            val, chain = ModelManager._dont_appear_in_child(a, link, [a])
            if not val:
                raise Exception(f"has circular links: {chain}")



import re
from itertools import combinations
from typing import Sequence, TypeVar


def name_convert_to_snake(name: str) -> str:
    return re.sub(r'([a-z])([A-Z])', r'\1_\2', name).lower()


def plural(word: str) -> str:
    rules = [
        (r'([bcdfghjklmnpqrstvwxz])y$', r'\1ies'),
        (r'(s|x|z|ch|sh)$', r'\1es'),
        (r'([^aeiou])o$', r'\1oes'),
        (r'$', r's')
    ]

    for pattern, replacement in rules:
        if re.search(pattern, word):
            return re.sub(pattern, replacement, word)

    return word


T = TypeVar('T')


def get_combinations(arr: Sequence[T]) -> Sequence[Sequence[T]]:
    result = []
    for r in range(1, len(arr) + 1):
        for combination in combinations(arr, r):
            result.append(combination)
    return result



from abc import ABC
from typing import List, Sequence, Optional
from enum import StrEnum

from fastapi_toolkit.generate.model_metadata import ModelMetadata, RelationshipMetadata, RelationshipSide
from fastapi_toolkit.generate.utils import get_combinations


class RouteMethod(StrEnum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'


class BaseRoute(ABC):
    methods: RouteMethod
    name: str
    url: str
    need_crud: bool = True


class CreateRoute(BaseRoute):
    methods = RouteMethod.POST

    def __init__(self, with_relation: Optional[RelationshipMetadata] = None,
                 link_to_route: Optional['RelationRoute'] = None):
        self.name = 'create_one'
        self.url = '/create_one'
        if with_relation is None:
            with_relation = []
        if with_relation:
            self.need_crud = False
            self.relation = with_relation
            self.name += '_with_' + with_relation.target.snake_name
            self.url += '_with_' + with_relation.target.snake_name
            self.link_to_route = link_to_route


class UpdateRoute(BaseRoute):
    methods = RouteMethod.PUT

    def __init__(self):
        self.name = 'update_one'
        self.url = '/update_one'


class QueryRoute(BaseRoute):
    methods = RouteMethod.GET
    schema_suffix = ''

    def __init__(self, is_all=False, with_relation: Sequence[RelationshipMetadata] = None):
        if with_relation is None:
            with_relation = []
        self.is_all = is_all
        self.name = 'get_all' if is_all else 'get_one'
        self.url = '/get_all' if is_all else '/get_one'
        if with_relation:
            self.name += '_with_' + '_'.join([r.target.snake_name for r in with_relation])
            self.url += '_with_' + '_'.join([r.target.snake_name for r in with_relation])
            self.schema_suffix = 'With' + 'And'.join([r.target.name for r in with_relation])


class DeleteRoute(BaseRoute):
    methods = RouteMethod.DELETE

    def __init__(self, is_all=False):
        self.is_all = is_all
        self.name = 'delete_all' if is_all else 'delete_one'
        self.url = '/delete_all' if is_all else '/delete_one'


class RelationRoute(BaseRoute):
    methods = RouteMethod.POST

    def __init__(self, relation: RelationshipMetadata, is_delete=False):
        self.relation = relation
        self.is_delete = is_delete
        self.name = f'link_to_' if not is_delete else f'unlink_to_'
        self.url = f'/link_to_' if not is_delete else f'/unlink_to_'
        self.name += relation.target.snake_name
        self.url += relation.target.snake_name


class RouterMetadata:
    model: ModelMetadata
    routes: List[BaseRoute]
    create_one: CreateRoute
    update_one: UpdateRoute
    get_one: QueryRoute
    get_all: QueryRoute
    delete_one: DeleteRoute
    delete_all: DeleteRoute

    def __init__(self, model: ModelMetadata):
        self.model = model
        self.create_one = CreateRoute()
        self.update_one = UpdateRoute()
        self.get_one = QueryRoute()
        self.get_all = QueryRoute(is_all=True)
        self.delete_one = DeleteRoute()
        self.delete_all = DeleteRoute(is_all=True)
        self.routes = [
            self.create_one,
            self.update_one,
            self.get_one,
            self.get_all,
            self.delete_one,
            self.delete_all,
        ]
        for relation in model.relationship:
            if len(relation.target.pk) > 1:
                print(f"Warning: {relation.target.name} has more than one pk")
                continue
            link_to_route = RelationRoute(relation)
            self.routes.append(link_to_route)
            self.routes.append(RelationRoute(relation, is_delete=True))
            if relation.side == RelationshipSide.one:
                self.routes.append(CreateRoute(with_relation=relation, link_to_route=link_to_route))
        for cbs in get_combinations(model.relationship):
            self.routes.append(QueryRoute(with_relation=cbs))
            self.routes.append(QueryRoute(is_all=True, with_relation=cbs))

    def query_routes(self):
        return [r for r in self.routes if isinstance(r, QueryRoute) and r.need_crud]

    def create_routes(self):
        return [r for r in self.routes if isinstance(r, CreateRoute) and r.need_crud]

    def update_routes(self):
        return [r for r in self.routes if isinstance(r, UpdateRoute) and r.need_crud]

    def delete_routes(self):
        return [r for r in self.routes if isinstance(r, DeleteRoute) and r.need_crud]

    def relation_routes(self):
        return [r for r in self.routes if isinstance(r, RelationRoute) and r.need_crud]

    def link_create_routes(self):
        return [r for r in self.routes if isinstance(r, CreateRoute) and not r.need_crud]



import importlib
import importlib.util as import_utils
import os
from pathlib import Path

import typer

from fastapi_toolkit.generate import CodeGenerator

app = typer.Typer()

tk_root = Path('.fastapi-toolkit')


def add_path_to_gitignore(path: Path):
    gitignore_path = Path('.gitignore')
    if not gitignore_path.exists():
        with open(gitignore_path, 'w') as f:
            f.write(f'{path}\n')
    else:
        with open(gitignore_path, 'r') as f:
            lines = f.readlines()
        if path.name not in lines:
            with open(gitignore_path, 'a') as f:
                f.write(f'{path}\n')


def import_module(module_name, module_path):
    spec = import_utils.spec_from_file_location(module_name, module_path)
    module = import_utils.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


@app.command('i')
@app.command('init')
def init():
    os.mkdir(tk_root)
    add_path_to_gitignore(tk_root)


@app.command('g')
@app.command('generate')
def generate(metadata_path: Path = 'metadata', root_path: Path = 'inner_code',
             table: bool = True, router: bool = True, mock: bool = True):
    if not root_path.is_dir():
        typer.confirm(f'root_path: {root_path} is not a dir, do you want to create it?', abort=True)
        root_path.mkdir(parents=True)
    module_name = "models"
    import_module(module_name, metadata_path.joinpath(f'{module_name}.py'))
    generator = CodeGenerator(root_path)
    generator.parse_models()
    if table:
        generator.generate_tables()
    if router:
        generator.generate_routers()
    if mock:
        generator.generate_mock()


@app.command('mock')
@app.command('m')
def mock(root_path: Path = 'inner_code'):
    main = importlib.import_module(str(root_path).replace('\\', '.') + '.mock').main
    main()


db_app = typer.Typer()


def get_dev_db(root_path: Path):
    module_path = root_path.joinpath('dev').joinpath(f'db.py')
    return import_module('db', module_path)


@db_app.command('init')
@db_app.command('i')
def db_init(root_path: Path = 'inner_code'):
    init = get_dev_db(root_path).init
    init(str(root_path).replace('\\', '.'))


@db_app.command('migrate')
@db_app.command('m')
def db_migrate(root_path: Path = 'inner_code', msg: str = None):
    migrate = get_dev_db(root_path).migrate
    migrate(msg)


@db_app.command('upgrade')
@db_app.command('u')
def db_upgrade(root_path: Path = 'inner_code'):
    upgrade = get_dev_db(root_path).upgrade
    upgrade()


app.add_typer(db_app, name="db")
